#ifndef lint
static	char sccsid[] = "@(#)mon.c 1.1 86/09/24 SMI";
#endif

/*
 *	Environment variable PROFDIR added such that:
 *		If PROFDIR doesn't exist, "mon.out" is produced as before.
 *		If PROFDIR = NULL, no profiling output is produced.
 *		If PROFDIR = string, "string/pid.progname" is produced,
 *		  where name consists of argv[0] suitably massaged.
 */
#include <sys/param.h>
#include "mon.h"
#define PROFDIR	"PROFDIR"

extern int creat(), write(), close(), getpid();
extern void profil(), perror();
extern char *getenv(), *strcpy(), *strrchr();

void monitor(), moncontrol();

char **___Argv = NULL; /* initialized to argv array by mcrt0 (if loaded) */
extern
struct cnt *countbase;

extern int numctrs;

extern int	profiling;
static char	*minsbrk = 0;
static char	*s_sbuf;
static int	s_bufsiz;
static int	s_scale;
static int	s_lowpc;

#define	MSG "No space for monitor buffer(s)\n"

static char mon_out[MAXPATHLEN];

monstartup(lowpc, highpc)
	char *lowpc;
	char *highpc;
{
	int monsize;
	char *buffer;
	int cntsiz;
	char *sbrk();

	cntsiz = (highpc - lowpc) * ARCDENSITY / 100;
	if (cntsiz < MINARCS)
		cntsiz = MINARCS;
	monsize = (highpc - lowpc + HISTFRACTION - 1) / HISTFRACTION
		+ sizeof(struct phdr) + cntsiz * sizeof(struct cnt);
	buffer = sbrk(monsize);
	if (buffer == (char *)-1) {
		write(2, MSG, sizeof(MSG));
		return;
	}
	minsbrk = sbrk(0);
	monitor(lowpc, highpc, buffer, monsize, cntsiz);
}

void
monitor(lowpc, highpc, buf, bufsiz, cntsiz)
	char *lowpc, *highpc;	/* boundaries of text to be monitored */
	char *buf;		/* ptr to space for monitor data (WORDs) */
	int bufsiz;		/* size of above space (in WORDs) */
	int cntsiz;		/* max no. of functions whose calls are counted */
{
	register int o;
	struct phdr *php;
	static int ssiz;
	static char *sbuf;
	register char *s, *end, *name = mon_out;

	if (lowpc == 0) {
		moncontrol(0);
		if (sbuf != 0) {
			if ((o = creat(mon_out, 0666)) < 0 ||
			    write(o, sbuf, (unsigned)ssiz) == -1)
				perror(mon_out);
			if (o >= 0)
				close(o);
		}
		return;
	}
	countbase = (struct cnt *)(buf + sizeof(struct phdr));
	sbuf = NULL;
	o = sizeof(struct phdr) + cntsiz * sizeof(struct cnt);
	if (ssiz >= bufsiz || lowpc >= highpc)
		return;		/* buffer too small or PC range bad */
	if ((s = getenv(PROFDIR)) == NULL) /* PROFDIR not in environment */
		end = MON_OUT; /* use default "mon.out" */
	else if (*s == '\0') /* value of PROFDIR is NULL */
		return; /* no profiling on this run */
	else { /* construct "PROFDIR/pid.progname" */
		register int pid, n;

		while (*s != '\0') /* copy PROFDIR value (path-prefix) */
			*name++ = *s++;
		*name++ = '/'; /* two slashes won't hurt */
		if ((pid = getpid()) <= 0) /* extra test just in case */
			pid = 1; /* getpid returns something inappropriate */
		for (n = 10000; n > pid; n /= 10)
			; /* suppress leading zeros */
		for ( ; ; n /= 10) {
			*name++ = pid/n + '0';
			if (n == 1)
				break;
			pid %= n;
		}
		*name++ = '.';
		if (___Argv != NULL) /* mcrt0.s executed */
			end = ((s = strrchr(___Argv[0], '/')) != NULL) ?
			    s + 1 : /* use only file-name part */
			    ___Argv[0]; /* use entire name */
	}
	(void) strcpy(name, end); /* copy final element of file-name */
	sbuf = buf;		/* for writing buffer at the wrapup */
	ssiz = bufsiz;
	php = (struct phdr *)&buf[0];
	php->lpc = (char *)lowpc;	/* initialize the first */
	php->hpc = (char *)highpc;	/* region of the buffer */
	php->ncnt = cntsiz;
	numctrs = cntsiz;
	buf += o;
	bufsiz -= o;
	if (bufsiz <= 0)
		return;
	o = (highpc - lowpc);
	if(bufsiz < o)
		o = ((float) bufsiz / o) * 65536;
	else
		o = 65536;
	s_scale = o;
	s_sbuf = buf;
	s_bufsiz = bufsiz;
	s_lowpc = (int) lowpc;
	moncontrol(1);
}

/*
 * Control profiling
 *	profiling is what mcount checks to see if
 *	all the data structures are ready.
 */
void
moncontrol(mode)
    int mode;
{
    if (mode) {
	/* start */
	profil(s_sbuf, s_bufsiz, s_lowpc, s_scale);
	profiling = 0;
    } else {
	/* stop */
	profil((char *)0, 0, 0, 0);
	profiling = 3;
    }
}
